@@ -33,6 +33,7 @@
 #include <libwindow-settings/gnome-wm-manager.h>
 #include <string.h>
 #include <libgnomevfs/gnome-vfs.h>
+#include <libgnomeui/gnome-thumbnail.h>
 
 #define CUSTOM_THEME_NAME "__custom__"
 
@@ -54,26 +55,113 @@ static const GtkTargetEntry drop_types[] =
   {"_NETSCAPE_URL", 0, TARGET_NS_URL}
 };
 
+static time_t
+theme_get_mtime (const char *name)
+{
+  GnomeThemeMetaInfo *theme;
+  time_t mtime = -1;
+
+  theme = gnome_theme_meta_info_find (name);
+  if (theme != NULL) {
+    GnomeVFSFileInfo *file_info;
+    GnomeVFSResult result;
+
+    file_info = gnome_vfs_file_info_new ();
+
+    result = gnome_vfs_get_file_info (theme->path, file_info,
+				      GNOME_VFS_FILE_INFO_DEFAULT |
+				      GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
+
+    if (result == GNOME_VFS_OK)
+      mtime = file_info->mtime;
+
+    gnome_vfs_file_info_unref (file_info);
+  }
+
+  return mtime;
+}
+
+static void
+theme_thumbnail_update (GdkPixbuf *pixbuf,
+			gchar *theme_name,
+			AppearanceData *data,
+			gboolean cache)
+{
+  GtkTreeIter iter;
+  GtkTreeModel *model = GTK_TREE_MODEL (data->theme_store);
+
+  /* find item in model and update thumbnail */
+  if (!pixbuf)
+    return;
+
+  if (theme_find_in_model (model, theme_name, &iter)) {
+    time_t mtime;
+
+    gtk_list_store_set (data->theme_store, &iter, COL_THUMBNAIL, pixbuf, -1);
+
+    /* cache thumbnail */
+    if (cache && (mtime = theme_get_mtime (theme_name)) != -1) {
+      gchar *path;
+
+      /* try to share thumbs with nautilus, use themes:/// */
+      path = g_strconcat ("themes:///", theme_name, NULL);
+
+      gnome_thumbnail_factory_save_thumbnail (data->thumb_factory,
+					      pixbuf, path, mtime);
+
+      g_free (path);
+    }
+  }
+
+  g_object_unref (pixbuf);
+}
+
+static GdkPixbuf *
+theme_get_thumbnail_from_cache (GnomeThemeMetaInfo *info, AppearanceData *data)
+{
+  GdkPixbuf *thumb = NULL;
+  gchar *path, *thumb_filename;
+  time_t mtime;
+
+  if (info == data->theme_custom)
+    return NULL;
+
+  mtime = theme_get_mtime (info->name);
+  if (mtime == -1)
+    return NULL;
+
+  /* try to share thumbs with nautilus, use themes:/// */
+  path = g_strconcat ("themes:///", info->name, NULL);
+  thumb_filename = gnome_thumbnail_factory_lookup (data->thumb_factory,
+                                                   path, mtime);
+  g_free (path);
+
+  if (thumb_filename != NULL) {
+    thumb = gdk_pixbuf_new_from_file (thumb_filename, NULL);
+    g_free (thumb_filename);
+  }
+
+  return thumb;
+}
+
 static void
 theme_thumbnail_done_cb (GdkPixbuf *pixbuf, gchar *theme_name, AppearanceData *data)
 {
-  /* find item in model and update thumbnail */
-  if (pixbuf) {
-    GtkTreeIter iter;
-    GtkTreeModel *model = GTK_TREE_MODEL (data->theme_store);
-
-    if (theme_find_in_model (model, theme_name, &iter))
-      gtk_list_store_set (data->theme_store, &iter, COL_THUMBNAIL, pixbuf, -1);
-
-    g_object_unref (pixbuf);
-  }
+  theme_thumbnail_update (pixbuf, theme_name, data, TRUE);
 }
 
 static void
 theme_thumbnail_generate (GnomeThemeMetaInfo *info, AppearanceData *data)
 {
-  generate_meta_theme_thumbnail_async (info,
-      (ThemeThumbnailFunc) theme_thumbnail_done_cb, data, NULL);
+  GdkPixbuf *thumb;
+
+  thumb = theme_get_thumbnail_from_cache (info, data);
+
+  if (thumb != NULL)
+    theme_thumbnail_update (thumb, info->name, data, FALSE);
+  else
+    generate_meta_theme_thumbnail_async (info,
+        (ThemeThumbnailFunc) theme_thumbnail_done_cb, data, NULL);
 }
 
 static void
@@ -123,6 +211,23 @@ get_default_string_from_key (GConfClient *client, const char *key)
   return str;
 }
 
+static gint
+get_default_int_from_key (GConfClient *client, const char *key, gint def)
+{
+  GConfValue *value;
+  gint val = def;
+
+  value = gconf_client_get_default_from_schema (client, key, NULL);
+
+  if (value) {
+    if (value->type == GCONF_VALUE_INT)
+      val = gconf_value_get_int (value);
+    gconf_value_free (value);
+  }
+
+  return val;
+}
+
 /* Find out if the lockdown key has been set. Currently returns false on error... */
 static gboolean
 is_locked_down (GConfClient *client)
@@ -154,13 +259,15 @@ theme_load_from_gconf (GConfClient *client, GnomeThemeMetaInfo *theme)
   theme->icon_theme_name = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
 
   g_free (theme->cursor_theme_name);
-
 #ifdef HAVE_XCURSOR
   theme->cursor_theme_name = gconf_client_get_string (client, CURSOR_THEME_KEY, NULL);
   theme->cursor_size = gconf_client_get_int (client, CURSOR_SIZE_KEY, NULL);
 #else
   theme->cursor_theme_name = gconf_client_get_string (client, CURSOR_FONT_KEY, NULL);
 #endif
+
+  g_free (theme->application_font);
+  theme->application_font = gconf_client_get_string (client, APPLICATION_FONT_KEY, NULL);
 }
 
 static gchar *
@@ -272,10 +379,14 @@ theme_set_custom_from_theme (const GnomeThemeMetaInfo *info, AppearanceData *dat
     g_free (custom->icon_theme_name);
     g_free (custom->metacity_theme_name);
     g_free (custom->gtk_color_scheme);
+    g_free (custom->cursor_theme_name);
+    g_free (custom->application_font);
     custom->gtk_theme_name = NULL;
     custom->icon_theme_name = NULL;
     custom->metacity_theme_name = NULL;
     custom->gtk_color_scheme = NULL;
+    custom->cursor_theme_name = NULL;
+    custom->application_font = NULL;
 
     if (info->gtk_theme_name)
       custom->gtk_theme_name = g_strdup (info->gtk_theme_name);
@@ -296,6 +407,25 @@ theme_set_custom_from_theme (const GnomeThemeMetaInfo *info, AppearanceData *dat
       custom->gtk_color_scheme = g_strdup (info->gtk_color_scheme);
     else
       custom->gtk_color_scheme = get_default_string_from_key (data->client, COLOR_SCHEME_KEY);
+
+    if (info->cursor_theme_name) {
+      custom->cursor_theme_name = g_strdup (info->cursor_theme_name);
+#ifdef HAVE_XCURSOR
+      custom->cursor_size = info->cursor_size;
+#endif
+    } else {
+#ifdef HAVE_XCURSOR
+      custom->cursor_theme_name = get_default_string_from_key (data->client, CURSOR_THEME_KEY);
+      custom->cursor_size = get_default_int_from_key (data->client, CURSOR_SIZE_KEY, 18);
+#else
+      custom->cursor_theme_name = get_default_string_from_key (data->client, CURSOR_FONT_KEY);
+#endif
+    }
+
+    if (info->application_font)
+      custom->application_font = g_strdup (info->application_font);
+    else
+      custom->application_font = get_default_string_from_key (data->client, APPLICATION_FONT_KEY);
   }
 
   /* select the custom theme */
@@ -607,9 +737,9 @@ theme_details_changed_cb (AppearanceData *data)
 }
 
 static void
-theme_color_scheme_changed_cb (GObject *settings,
-                               GParamSpec *pspec,
-                               AppearanceData *data)
+theme_setting_changed_cb (GObject *settings,
+                          GParamSpec *pspec,
+                          AppearanceData *data)
 {
   theme_details_changed_cb (data);
 }
@@ -648,10 +778,17 @@ theme_store_sort_func (GtkTreeModel *model,
 
     gtk_tree_model_get (model, b, COL_NAME, &b_name, COL_LABEL, &b_label, -1);
 
-    if (!strcmp (b_name, CUSTOM_THEME_NAME))
+    if (!strcmp (b_name, CUSTOM_THEME_NAME)) {
       rc = 1;
-    else
-      rc = strcmp (a_label, b_label);
+    } else {
+      gchar *a_case, *b_case;
+
+      a_case = g_utf8_casefold (a_label, -1);
+      b_case = g_utf8_casefold (b_label, -1);
+      rc = g_utf8_collate (a_case, b_case);
+      g_free (a_case);
+      g_free (b_case);
+    }
 
     g_free (b_name);
     g_free (b_label);
@@ -664,15 +801,14 @@ theme_store_sort_func (GtkTreeModel *model,
 }
 
 static void
-appearance_window_drag_data_received_cb (GtkWidget *widget,
-                                         GdkDragContext *context,
-                                         gint x, gint y,
-                                         GtkSelectionData *selection_data,
-                                         guint info, guint time,
-                                         AppearanceData *data)
+theme_drag_data_received_cb (GtkWidget *widget,
+                             GdkDragContext *context,
+                             gint x, gint y,
+                             GtkSelectionData *selection_data,
+                             guint info, guint time,
+                             AppearanceData *data)
 {
   GList *uris;
-  gchar *filename = NULL;
 
   if (!(info == TARGET_URI_LIST || info == TARGET_NS_URL))
     return;
@@ -681,6 +817,7 @@ appearance_window_drag_data_received_cb (GtkWidget *widget,
 
   if (uris != NULL && uris->data != NULL) {
     GnomeVFSURI *uri = (GnomeVFSURI *) uris->data;
+    gchar *filename;
 
     if (gnome_vfs_uri_is_local (uri))
       filename = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (uri), G_DIR_SEPARATOR_S);
@@ -688,10 +825,11 @@ appearance_window_drag_data_received_cb (GtkWidget *widget,
       filename = gnome_vfs_unescape_string (gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE), G_DIR_SEPARATOR_S);
 
     gnome_vfs_uri_list_unref (uris);
-  }
 
-  gnome_theme_install_from_uri (filename, GTK_WINDOW (widget));
-  g_free (filename);
+    gnome_theme_install_from_uri (filename,
+        GTK_WINDOW (glade_xml_get_widget (data->xml, "appearance_window")));
+    g_free (filename);
+  }
 }
 
 static void
@@ -713,6 +851,7 @@ themes_init (AppearanceData *data)
   GnomeThemeMetaInfo *meta_theme = NULL;
   GtkIconView *icon_view;
   GtkCellRenderer *renderer;
+  GtkSettings *settings;
 
   /* initialise some stuff */
   gnome_theme_init (NULL);
@@ -725,8 +864,6 @@ themes_init (AppearanceData *data)
   data->theme_store = theme_store =
       gtk_list_store_new (NUM_COLS, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING);
 
-  del_button = glade_xml_get_widget (data->xml, "theme_delete");
-
   /* set up theme list */
   theme_list = gnome_theme_meta_info_find_all ();
   gnome_theme_info_register_theme_change ((ThemeChangedCallback) theme_changed_on_disk_cb, data);
@@ -766,12 +903,6 @@ themes_init (AppearanceData *data)
   g_list_foreach (theme_list, (GFunc) theme_thumbnail_generate, data);
   g_list_free (theme_list);
 
-  w = glade_xml_get_widget (data->xml, "appearance_window");
-  gtk_drag_dest_set (w, GTK_DEST_DEFAULT_ALL,
-		     drop_types, G_N_ELEMENTS (drop_types),
-		     GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
-  g_signal_connect (w, "drag-data-received", (GCallback) appearance_window_drag_data_received_cb, NULL);
-
   icon_view = GTK_ICON_VIEW (glade_xml_get_widget (data->xml, "theme_list"));
 
   renderer = cell_renderer_wallpaper_new ();
@@ -784,7 +915,9 @@ themes_init (AppearanceData *data)
   renderer = gtk_cell_renderer_text_new ();
   g_object_set (renderer, "alignment", PANGO_ALIGN_CENTER,
 			  "wrap-mode", PANGO_WRAP_WORD_CHAR,
-			  "xalign", 0.5, "yalign", 0.0, NULL);
+			  "wrap-width", gtk_icon_view_get_item_width (icon_view),
+			  "width", gtk_icon_view_get_item_width (icon_view),
+			  "xalign", 0.0, "yalign", 0.0, NULL);
   gtk_cell_layout_pack_end (GTK_CELL_LAYOUT (icon_view), renderer, FALSE);
   gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (icon_view), renderer,
                                   "markup", COL_LABEL, NULL);
@@ -795,7 +928,7 @@ themes_init (AppearanceData *data)
   gtk_icon_view_set_model (icon_view, GTK_TREE_MODEL (sort_model));
 
   g_signal_connect (icon_view, "selection-changed", (GCallback) theme_selection_changed_cb, data);
-  theme_select_name (icon_view, meta_theme->name);
+  g_signal_connect_after (icon_view, "realize", (GCallback) theme_select_name, meta_theme->name);
 
   w = glade_xml_get_widget (data->xml, "theme_install");
   gtk_button_set_image (GTK_BUTTON (w),
@@ -812,24 +945,30 @@ themes_init (AppearanceData *data)
                         gtk_image_new_from_stock (GTK_STOCK_EDIT, GTK_ICON_SIZE_BUTTON));
   g_signal_connect (w, "clicked", (GCallback) theme_custom_cb, data);
 
+  del_button = glade_xml_get_widget (data->xml, "theme_delete");
   g_signal_connect (del_button, "clicked", (GCallback) theme_delete_cb, data);
 
+  w = glade_xml_get_widget (data->xml, "theme_vbox");
+  gtk_drag_dest_set (w, GTK_DEST_DEFAULT_ALL,
+		     drop_types, G_N_ELEMENTS (drop_types),
+		     GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
+  g_signal_connect (w, "drag-data-received", (GCallback) theme_drag_data_received_cb, data);
+  if (is_locked_down (data->client))
+    gtk_widget_set_sensitive (w, FALSE);
+
   /* listen to gconf changes, too */
   gconf_client_add_dir (data->client, "/apps/metacity/general", GCONF_CLIENT_PRELOAD_NONE, NULL);
   gconf_client_add_dir (data->client, "/desktop/gnome/interface", GCONF_CLIENT_PRELOAD_NONE, NULL);
-  gconf_client_notify_add (data->client, GTK_THEME_KEY, (GConfClientNotifyFunc) theme_gconf_changed, data, NULL, NULL);
   gconf_client_notify_add (data->client, METACITY_THEME_KEY, (GConfClientNotifyFunc) theme_gconf_changed, data, NULL, NULL);
-  gconf_client_notify_add (data->client, ICON_THEME_KEY, (GConfClientNotifyFunc) theme_gconf_changed, data, NULL, NULL);
   gconf_client_notify_add (data->client, BACKGROUND_KEY, (GConfClientNotifyFunc) background_or_font_changed, data, NULL, NULL);
   gconf_client_notify_add (data->client, APPLICATION_FONT_KEY, (GConfClientNotifyFunc) background_or_font_changed, data, NULL, NULL);
   gconf_client_notify_add (data->client, DESKTOP_FONT_KEY, (GConfClientNotifyFunc) background_or_font_changed, data, NULL, NULL);
   gconf_client_notify_add (data->client, MONOSPACE_FONT_KEY, (GConfClientNotifyFunc) background_or_font_changed, data, NULL, NULL);
 
-  g_signal_connect (gtk_settings_get_default (), "notify::gtk-color-scheme", (GCallback) theme_color_scheme_changed_cb, data);
-
-  if (is_locked_down (data->client)) {
-    /* FIXME: determine what needs disabling */
-  }
+  settings = gtk_settings_get_default ();
+  g_signal_connect (settings, "notify::gtk-color-scheme", (GCallback) theme_setting_changed_cb, data);
+  g_signal_connect (settings, "notify::gtk-theme-name", (GCallback) theme_setting_changed_cb, data);
+  g_signal_connect (settings, "notify::gtk-icon-theme-name", (GCallback) theme_setting_changed_cb, data);
 }
 
 void
