@@ -1404,12 +1404,19 @@ gnome_theme_info_find_by_type_helper (gpointer key,
 				      struct GnomeThemeInfoHashData *hash_data)
 {
   guint elements = GPOINTER_TO_INT (hash_data->user_data);
-  GnomeThemeInfo *theme_info = list->data;
 
-  if ((elements & GNOME_THEME_METACITY && theme_info->has_metacity) ||
-      (elements & GNOME_THEME_GTK_2 && theme_info->has_gtk) ||
-      (elements & GNOME_THEME_GTK_2_KEYBINDING && theme_info->has_keybinding))
-    hash_data->list = g_list_prepend (hash_data->list, theme_info);
+  do {
+    GnomeThemeInfo *theme_info = list->data;
+
+    if ((elements & GNOME_THEME_METACITY && theme_info->has_metacity) ||
+        (elements & GNOME_THEME_GTK_2 && theme_info->has_gtk) ||
+        (elements & GNOME_THEME_GTK_2_KEYBINDING && theme_info->has_keybinding)) {
+      hash_data->list = g_list_prepend (hash_data->list, theme_info);
+      return;
+    }
+
+    list = list->next;
+  } while (list);
 }
 
 GList *
@@ -1703,7 +1710,7 @@ gnome_theme_is_writable (const gpointer theme, GnomeThemeType type) {
       theme_path = ((const GnomeThemeIconInfo *) theme)->path;
       break;
     case GNOME_THEME_TYPE_CURSOR:
-      theme_path = ((const GnomeThemeIconInfo *) theme)->path;
+      theme_path = ((const GnomeThemeCursorInfo *) theme)->path;
       break;
     case GNOME_THEME_TYPE_METATHEME:
       theme_path = ((const GnomeThemeMetaInfo *) theme)->path;
@@ -1796,10 +1803,10 @@ read_current_cursor_font (void)
 static void
 read_cursor_fonts (void)
 {
-  gchar* cursor_font;
+  gchar *cursor_font;
   gint i;
 
-  static const gchar* builtins[][4] = {
+  const gchar *builtins[][4] = {
     {
       "gnome/cursor-fonts/cursor-normal.pcf",
       N_("Default Pointer"),
@@ -1829,17 +1836,17 @@ read_cursor_fonts (void)
     cursor_font = g_strdup (builtins[0][0]);
 
   for (i = 0; i < G_N_ELEMENTS (builtins); i++) {
-		GnomeThemeCursorInfo *theme_info;
-		gchar *filename;
+    GnomeThemeCursorInfo *theme_info;
+    gchar *filename;
 
     theme_info = gnome_theme_cursor_info_new ();
-    theme_info->priority = 0;
 
     filename = g_build_filename (GNOMECC_DATA_DIR "/pixmaps", builtins[i][3], NULL);
     theme_info->thumbnail = gdk_pixbuf_new_from_file (filename, NULL);
     g_free (filename);
 
-    theme_info->name = theme_info->path = g_build_filename (GNOMECC_DATA_DIR, builtins[i][0], NULL);
+    theme_info->name = g_build_filename (GNOMECC_DATA_DIR, builtins[i][0], NULL);
+    theme_info->path = g_strdup (theme_info->name);
 
     if (!strcmp (theme_info->path, cursor_font))
       theme_info->readable_name = g_strdup (builtins[i][2]);
@@ -1849,6 +1856,8 @@ read_cursor_fonts (void)
     g_hash_table_insert (cursor_theme_hash_by_uri, theme_info->path, theme_info);
     add_data_to_hash_by_name (cursor_theme_hash_by_name, theme_info->name, theme_info);
   }
+
+  g_free (cursor_font);
 }
 #endif /* !HAVE_XCURSOR */
 
@@ -1941,8 +1950,8 @@ gnome_theme_init (gboolean *monitor_not_added)
   meta_theme_hash_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
   icon_theme_hash_by_uri = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
   icon_theme_hash_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
-  cursor_theme_hash_by_uri = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
-  cursor_theme_hash_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+  cursor_theme_hash_by_uri = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
+  cursor_theme_hash_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
   theme_hash_by_uri = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
   theme_hash_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
@@ -1981,6 +1990,18 @@ gnome_theme_init (gboolean *monitor_not_added)
     {
       top_theme_dir_string = g_build_filename (INSTALL_PREFIX, "share", "icons", NULL);
     }
+
+#ifdef XCURSOR_ICONDIR
+  /* if there's a separate xcursors dir, add that as well */
+  if (strcmp (XCURSOR_ICONDIR, top_theme_dir_string) &&
+      strcmp (XCURSOR_ICONDIR, "/usr/share/icons")) {
+    top_theme_dir_uri = gnome_vfs_uri_new (XCURSOR_ICONDIR);
+    if (gnome_vfs_uri_exists (top_theme_dir_uri))
+      result = add_top_icon_theme_dir_monitor (top_theme_dir_uri, &real_monitor_not_added, 1, NULL);
+    gnome_vfs_uri_unref (top_theme_dir_uri);
+  }
+#endif
+
   top_theme_dir_uri = gnome_vfs_uri_new (top_theme_dir_string);
   g_free (top_theme_dir_string);
 
