@@ -16,6 +16,7 @@
 #include <libgnomeui/gnome-icon-theme.h>
 
 #include "theme-thumbnail.h"
+#include "gtkrc-utils.h"
 #include "capplet-util.h"
 
 static gint child_pid;
@@ -88,7 +89,7 @@ static int pipe_from_factory_fd[2];
 #define THUMBNAIL_TYPE_METACITY "metacity"
 #define THUMBNAIL_TYPE_ICON     "icon"
 
-#define META_THUMBNAIL_SIZE       140
+#define META_THUMBNAIL_SIZE       128
 #define GTK_THUMBNAIL_SIZE         96
 #define METACITY_THUMBNAIL_WIDTH  120
 #define METACITY_THUMBNAIL_HEIGHT  60
@@ -134,8 +135,6 @@ create_folder_icon (char *icon_theme_name)
   GtkIconInfo *folder_icon_info;
   const gchar *filename;
   gchar *example_icon_name;
-  int width, height;
-  gdouble scale;
 
   icon_theme = gtk_icon_theme_new ();
   gtk_icon_theme_set_custom_theme (icon_theme, icon_theme_name);
@@ -188,12 +187,19 @@ create_folder_icon (char *icon_theme_name)
    * them. The width is usually the largest dimension for icons that come at
    * irregular sizes, so use this to calculate the scale factor
    */
-  scale = ((double) ICON_THUMBNAIL_SIZE) / gdk_pixbuf_get_width (folder_icon);
-  width = 48;
-  height = scale * gdk_pixbuf_get_height (folder_icon);
+  if (gdk_pixbuf_get_width (folder_icon) != ICON_THUMBNAIL_SIZE) {
+    int width, height;
+    gdouble scale;
 
-  retval = gdk_pixbuf_scale_simple (folder_icon, width, height, GDK_INTERP_BILINEAR);
-  g_object_unref (folder_icon);
+    scale = ((double) ICON_THUMBNAIL_SIZE) / gdk_pixbuf_get_width (folder_icon);
+    width = ICON_THUMBNAIL_SIZE;
+    height = scale * gdk_pixbuf_get_height (folder_icon);
+
+    retval = gdk_pixbuf_scale_simple (folder_icon, width, height, GDK_INTERP_BILINEAR);
+    g_object_unref (folder_icon);
+  } else {
+    retval = folder_icon;
+  }
 
   return retval;
 }
@@ -215,13 +221,11 @@ create_meta_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
   GdkPixmap *pixmap;
   GdkVisual *visual;
   MetaFrameFlags flags;
-  MetaTheme *theme = NULL;
-  GtkSettings *settings;
+  MetaTheme *theme;
   GdkPixbuf *pixbuf, *icon;
-  int width, height;
+  int icon_width, icon_height;
 
-  settings = gtk_settings_get_default ();
-  g_object_set (G_OBJECT (settings),
+  g_object_set (gtk_settings_get_default (),
     "gtk-theme-name", (char *) theme_thumbnail_data->control_theme_name->data,
     "gtk-font-name", (char *) theme_thumbnail_data->application_font->data,
     "gtk-icon-theme-name", (char *) theme_thumbnail_data->icon_theme_name->data,
@@ -229,6 +233,12 @@ create_meta_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
     NULL);
   theme = meta_theme_load ((char *) theme_thumbnail_data->wm_theme_name->data, NULL);
 
+  /* Represent the icon theme */
+  icon = create_folder_icon ((char *) theme_thumbnail_data->icon_theme_name->data);
+  icon_width = gdk_pixbuf_get_width (icon);
+  icon_height = gdk_pixbuf_get_height (icon);
+
+  /* Create a fake window */
   flags = META_FRAME_ALLOWS_DELETE |
           META_FRAME_ALLOWS_MENU |
           META_FRAME_ALLOWS_MINIMIZE |
@@ -309,22 +319,14 @@ create_meta_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
   pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, META_THUMBNAIL_SIZE, META_THUMBNAIL_SIZE);
   gdk_pixbuf_get_from_drawable (pixbuf, pixmap, NULL, 0, 0, 0, 0, META_THUMBNAIL_SIZE, META_THUMBNAIL_SIZE);
 
-  /* Handle the icon theme */
-  icon = create_folder_icon ((char *) theme_thumbnail_data->icon_theme_name->data);
-  width = gdk_pixbuf_get_width (icon);
-  height = gdk_pixbuf_get_height (icon);
-
-  gdk_pixbuf_composite (icon,
-                        pixbuf,
-                        vbox->allocation.x + vbox->allocation.width - width - 5,
-                        vbox->allocation.y + vbox->allocation.height - height - 5,
-                        width,
-                        height,
-                        vbox->allocation.x + vbox->allocation.width - width - 5,
-                        vbox->allocation.y + vbox->allocation.height - height - 5,
-                        1.0,
-                        1.0,
-                        GDK_INTERP_BILINEAR, 255);
+  /* Add the icon theme to the pixbuf */
+  gdk_pixbuf_composite (icon, pixbuf,
+                        vbox->allocation.x + vbox->allocation.width - icon_width - 5,
+                        vbox->allocation.y + vbox->allocation.height - icon_height - 5,
+                        icon_width, icon_height,
+                        vbox->allocation.x + vbox->allocation.width - icon_width - 5,
+                        vbox->allocation.y + vbox->allocation.height - icon_height - 5,
+                        1.0, 1.0, GDK_INTERP_BILINEAR, 255);
   g_object_unref (icon);
   gtk_widget_destroy (window);
 
@@ -334,7 +336,7 @@ create_meta_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
 static GdkPixbuf *
 create_gtk_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
 {
-  GObject *settings;
+  GtkSettings *settings;
   GtkWidget *window, *vbox, *box, *stock_button, *checkbox, *radio;
   GtkRequisition requisition;
   GtkAllocation allocation;
@@ -343,8 +345,10 @@ create_gtk_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
   GdkPixbuf *pixbuf, *retval;
   gint width, height;
 
-  settings = G_OBJECT (gtk_settings_get_default ());
-  g_object_set (settings, "gtk-theme-name", (char *) theme_thumbnail_data->control_theme_name->data, NULL);
+  settings = gtk_settings_get_default ();
+  g_object_set (settings, "gtk-theme-name", (char *) theme_thumbnail_data->control_theme_name->data,
+			  "gtk-color-scheme", (char *) theme_thumbnail_data->gtk_color_scheme->data,
+ 			  NULL);
 
   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 
@@ -486,6 +490,7 @@ create_metacity_theme_pixbuf (ThemeThumbnailData *theme_thumbnail_data)
   g_object_unref (pixbuf);
 
   gtk_widget_destroy (window);
+  meta_theme_free (theme);
   return retval;
 }
 
@@ -511,8 +516,9 @@ handle_bytes (const gchar        *buffer,
     switch (theme_thumbnail_data->status)
     {
       case READY_FOR_THEME:
-      case READING_TYPE:
         theme_thumbnail_data->status = READING_TYPE;
+        /* fall through */
+      case READING_TYPE:
         nil = memchr (ptr, '\000', bytes_read);
         if (nil == NULL)
         {
@@ -624,7 +630,7 @@ message_from_capplet (GIOChannel   *source,
   gsize bytes_read;
   GdkPixbuf *pixbuf;
   gint i, rowstride;
-  char *pixels;
+  guchar *pixels;
   ThemeThumbnailData *theme_thumbnail_data;
 
   theme_thumbnail_data = (ThemeThumbnailData *) data;
@@ -654,12 +660,12 @@ message_from_capplet (GIOChannel   *source,
         else if (!strcmp (type, THUMBNAIL_TYPE_ICON))
           pixbuf = create_icon_theme_pixbuf (theme_thumbnail_data);
         else
-          pixbuf = NULL; /* should never happen */
+          g_assert_not_reached ();
 
         width = gdk_pixbuf_get_width (pixbuf);
         height = gdk_pixbuf_get_height (pixbuf);
         rowstride = gdk_pixbuf_get_rowstride (pixbuf);
-        pixels = (char *) gdk_pixbuf_get_pixels (pixbuf);
+        pixels = gdk_pixbuf_get_pixels (pixbuf);
 
         /* Write the pixbuf's size */
         write (pipe_from_factory_fd[1], &width, sizeof (width));
@@ -670,7 +676,7 @@ message_from_capplet (GIOChannel   *source,
           write (pipe_from_factory_fd[1], pixels + (rowstride)*i, width * gdk_pixbuf_get_n_channels (pixbuf));
         }
         g_object_unref (pixbuf);
-        theme_thumbnail_data->status = READY_FOR_THEME;;
+        theme_thumbnail_data->status = READY_FOR_THEME;
         g_byte_array_set_size (theme_thumbnail_data->type, 0);
         g_byte_array_set_size (theme_thumbnail_data->control_theme_name, 0);
         g_byte_array_set_size (theme_thumbnail_data->gtk_color_scheme, 0);
@@ -753,13 +759,9 @@ message_from_child (GIOChannel   *source,
 
       if (async_data.thumbnail_width == 0 && async_data.data->len >= 2 * sizeof (gint))
       {
-        int i;
-
         async_data.thumbnail_width = *((gint *) async_data.data->data);
         async_data.thumbnail_height = *(((gint *) async_data.data->data) + 1);
-
-        for (i = 0; i < 2 * sizeof (gint); i++)
-          g_byte_array_remove_index (async_data.data, 0);
+        g_byte_array_remove_range (async_data.data, 0, 2 * sizeof (gint));
       }
       else if (async_data.thumbnail_width > 0 && async_data.data->len == async_data.thumbnail_width * async_data.thumbnail_height * 4)
       {
@@ -858,7 +860,7 @@ read_pixbuf (void)
   gint size[2];
   GdkPixbuf *pixbuf;
   gint rowstride;
-  char *pixels;
+  guchar *pixels;
 
   do
   {
@@ -869,7 +871,7 @@ read_pixbuf (void)
 
   pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, size[0], size[1]);
   rowstride = gdk_pixbuf_get_rowstride (pixbuf);
-  pixels = (char *) gdk_pixbuf_get_pixels (pixbuf);
+  pixels = gdk_pixbuf_get_pixels (pixbuf);
 
   for (i = 0; i < size[1]; i++)
   {
@@ -906,8 +908,6 @@ generate_theme_thumbnail (gchar *thumbnail_type,
                           gchar *icon_theme_name,
                           gchar *application_font)
 {
-  GdkPixbuf *pixbuf;
-
   if (async_data.set || !pipe_to_factory_fd[1] || !pipe_from_factory_fd[0])
     return NULL;
 
@@ -918,8 +918,7 @@ generate_theme_thumbnail (gchar *thumbnail_type,
                           icon_theme_name,
                           application_font);
 
-  pixbuf = read_pixbuf ();
-  return pixbuf;
+  return read_pixbuf ();
 }
 
 GdkPixbuf *
@@ -947,7 +946,7 @@ generate_gtk_theme_thumbnail (GnomeThemeInfo *theme_info)
 GdkPixbuf *
 generate_metacity_theme_thumbnail (GnomeThemeInfo *theme_info)
 {
-  return generate_theme_thumbnail (THUMBNAIL_TYPE_GTK,
+  return generate_theme_thumbnail (THUMBNAIL_TYPE_METACITY,
                                    NULL,
                                    NULL,
                                    theme_info->name,
@@ -958,7 +957,7 @@ generate_metacity_theme_thumbnail (GnomeThemeInfo *theme_info)
 GdkPixbuf *
 generate_icon_theme_thumbnail (GnomeThemeIconInfo *theme_info)
 {
-  return generate_theme_thumbnail (THUMBNAIL_TYPE_GTK,
+  return generate_theme_thumbnail (THUMBNAIL_TYPE_ICON,
                                    NULL,
                                    NULL,
                                    NULL,
@@ -979,6 +978,8 @@ generate_theme_thumbnail_async (gpointer            theme_info,
                                 gpointer            user_data,
                                 GDestroyNotify      destroy)
 {
+  gchar *scheme;
+
   if (async_data.set)
   {
     ThemeQueueItem *item;
@@ -1004,7 +1005,6 @@ generate_theme_thumbnail_async (gpointer            theme_info,
     return;
   }
 
-
   if (async_data.channel == NULL)
   {
     async_data.channel = g_io_channel_unix_new (pipe_from_factory_fd[0]);
@@ -1021,12 +1021,18 @@ generate_theme_thumbnail_async (gpointer            theme_info,
   async_data.user_data = user_data;
   async_data.destroy = destroy;
 
+  if (!strcmp (thumbnail_type, THUMBNAIL_TYPE_GTK)) {
+    scheme = gtkrc_get_color_scheme_for_theme (theme_name);
+    gtk_color_scheme = scheme;
+  } else scheme = NULL;
+
   send_thumbnail_request (thumbnail_type,
                           gtk_theme_name,
                           gtk_color_scheme,
                           metacity_theme_name,
                           icon_theme_name,
                           application_font);
+  g_free (scheme);
 }
 
 void
@@ -1035,7 +1041,7 @@ generate_meta_theme_thumbnail_async (GnomeThemeMetaInfo *theme_info,
                                      gpointer            user_data,
                                      GDestroyNotify      destroy)
 {
-  return generate_theme_thumbnail_async (theme_info,
+  generate_theme_thumbnail_async (theme_info,
                                          theme_info->name,
                                          THUMBNAIL_TYPE_META,
                                          theme_info->gtk_theme_name,
@@ -1052,7 +1058,7 @@ generate_gtk_theme_thumbnail_async (GnomeThemeInfo *theme_info,
                                     gpointer            user_data,
                                     GDestroyNotify      destroy)
 {
-  return generate_theme_thumbnail_async (theme_info,
+  generate_theme_thumbnail_async (theme_info,
                                          theme_info->name,
                                          THUMBNAIL_TYPE_GTK,
                                          theme_info->name,
@@ -1069,7 +1075,7 @@ generate_metacity_theme_thumbnail_async (GnomeThemeInfo *theme_info,
                                          gpointer            user_data,
                                          GDestroyNotify      destroy)
 {
-  return generate_theme_thumbnail_async (theme_info,
+  generate_theme_thumbnail_async (theme_info,
                                          theme_info->name,
                                          THUMBNAIL_TYPE_METACITY,
                                          NULL,
@@ -1086,7 +1092,7 @@ generate_icon_theme_thumbnail_async (GnomeThemeIconInfo *theme_info,
                                      gpointer            user_data,
                                      GDestroyNotify      destroy)
 {
-  return generate_theme_thumbnail_async (theme_info,
+  generate_theme_thumbnail_async (theme_info,
                                          theme_info->name,
                                          THUMBNAIL_TYPE_ICON,
                                          NULL,
